
import numpy as np
from multiprocessing.shared_memory import SharedMemory
from multiprocessing import Lock

from ptlib.core.job import JobSpec, Job


class BaseQueue:
    """ Queue. *** COME BACK *** """

    class Wait:
        """ Flag for lock acquisition wait. """
    class Empty:
        """ Flag for emtpy queue. """

    class Full:
        """ Flag for full queue. """

    class Closed:
        """ Flag for closed queue. """

    def __init__(self, *args, **kwargs):
        self.job = None

    def get(self):
        return []

    def put(self):
        return True

    def close(self):
        pass

    def _link_mem(self, create_local=False):
        pass


class FIFOQueue(BaseQueue):
    """ Object to faciliate memory management across parallel processes.

    Attributes: (** update with jobs and replace arr with buffer or buf)
        _job_buffer -- ptlib.core.job.Job or (capacity x shape)-darray
            Array of arrays linked to the shared memory buffer for asynchronous
            data transfer.
        _local_job_buffer -- ptlib.core.job.Job or shape-darray
            Local copy of a single job used to load a temporary copy of an
            _job_buffer[i]. This is only created if `_link_mem` is called with
            kwarg `create_local=True`. Very important that this is created if
            queue is acting as an input or else data in `_job_buffer` might get
            overwritten by an upstream task before it is used.
        _arr_sel -- (3 x 1)-darray
            Buffer with following elements:
                =+= `arr_sel[0]=i` indicates that `put()` will store data in
                    `arr_dat[i]`
                =+= `arr_sel[1]=j` indicates that `get()` will retreive data
                    from `arr_dat[j]`.
                =+= `arr_sel[0]=k` where `k==0` indicates queue is open and
                    `k==1` indicates queue is closed.
        _arr_chk -- (capacity x 1)-darray
            Buffer where each `arr_chk[i]` is 0 if `arr_dat[i]` is full or 1 if
            `arr_dat[i]` is empty.
        """

    def __init__(self,
                 capacity: int,
                 job_spec: JobSpec):
        """
        Parameters:
            capacity -- int
                The number of data-sized slots to send data.
            job_spec -- ptlib.core.job.JobSpec
                Specification for the structure of the input job into the queue.
        """

        self.capacity = capacity
        self.job_spec = job_spec

        # define local job buffer
        self._job_buffer = None

        # create shared memory objects
        self._shm_dat = SharedMemory(
            create=True, size=job_spec.get_nbytes(capacity))
        self._shm_sel = SharedMemory(create=True, size=3)
        self._shm_chk = SharedMemory(create=True, size=capacity)

        # initialize selection and check arrays
        np.ndarray(2, dtype=np.int8, buffer=self._shm_sel.buf).fill(0)
        np.ndarray(capacity, dtype=np.int8, buffer=self._shm_chk.buf).fill(0)

        # define arrays
        self._arr_dat = np.ndarray(0)
        self._arr_sel = np.ndarray(0)
        self._arr_chk = np.ndarray(0)

        # create selection lock
        self._lock_sel = Lock()

        # flag to check if arrays have been linked in current context
        self._is_linked = False

    def get(self):
        """
        Attemps to retreive data from shared memory. If the selected index is 
        empty, this returns `BaseQueue.Empty`. If the queue is closed, then 
        `BaseQueue.Closed` is returned. Otherwise, this loads data into 
        `self._arr_dat` and returns True. 
        """

        # acquire selection lock
        self._lock_sel.acquire()

        # get index
        sel_index = self._arr_sel[1]

        # if get index == set index and check[set] is low, wait for check[set] to be high
        if self._arr_chk[sel_index] == 0:  # queue is either empty or closed
            self._lock_sel.release()
            # print(f"sel get: {sel_index} returning False", self._arr_chk[sel_index])

            return BaseQueue.Empty if self._arr_sel[2] == 0 else BaseQueue.Closed

        # increment get index
        self._arr_sel[1] = (sel_index + 1) % self.capacity
        # print(f"sel get: {sel_index}", self._arr_chk[sel_index])

        # get payload (must copy because buffer might change in other process)
        self._job_buffer[:] = self._arr_dat[sel_index]

        # release selection lock
        self._lock_sel.release()

        # set check[get] to low
        self._arr_chk[sel_index] = 0

        return True

    def put(self):
        """
        Attempts to load `self._job_buffer` into shared memory. If the selected 
        index is full, then this returns `BaseQueue.Full`. If the queue is 
        closed, then `BaseQueue.Closed` is returned. Otherwise, this loads data 
        into `self._arr_dat` and returns True. 
        """

        # acquire selection lock
        self._lock_sel.acquire()

        # set index
        sel_index = self._arr_sel[0]

        # if set index == get index and check[get] is high, wait for check[get] to be low
        if self._arr_chk[sel_index] == 1:  # queue is either full or closed
            self._lock_sel.release()
            # print(f"sel put: {sel_index} returning False", self._arr_chk[sel_index])

            return BaseQueue.Full if self._arr_sel[2] == 0 else BaseQueue.Closed

        # increment set index
        self._arr_sel[0] = (sel_index + 1) % self.capacity
        # print(f"sel put: {sel_index}", self._arr_chk[sel_index])

        # print(self._arr_dat, self._job_buffer)
        self._arr_dat[sel_index][:] = self._job_buffer

        # release selection lock
        self._lock_sel.release()

        # set check[set] to high
        self._arr_chk[sel_index] = 1

        return True

    def close(self):
        """
        Closes queue by setting `self._arr_sel[2]` to HIGH.
        """

        # link selection array if it isn't already
        if not self._is_linked:
            self._arr_sel = np.ndarray(
                3, dtype=np.int8, buffer=self._shm_sel.buf)

        # set `self._arr_sel[2]` to HIGH
        self._arr_sel[2] = 1

    def _link_mem(self):

        # create local array connected to shared memory buffer
        self._arr_dat = np.ndarray((self.capacity, self.job_spec.get_nbytes()),
                                   dtype=np.int8,
                                   buffer=self._shm_dat.buf)

        # create local job buffer
        local_job = Job(self.job_spec)
        self._job_buffer = local_job._buffer

        # link selection and check arrays to buffers in memory
        self._arr_sel = np.ndarray(
            3, dtype=np.int8, buffer=self._shm_sel.buf)
        self._arr_chk = np.ndarray(
            self.capacity, dtype=np.int8, buffer=self._shm_chk.buf)

        # set linked flag to HIGH
        self._is_linked = True

        return local_job

    # def _generate_dynamic_put(self):
    #     print("generating dynamic put")
    #     f = open("_dptest.py", "w")
    #     f.write("def _put(_job_buffer, buffer, sel_index):\n")
    #     for i in self._iter:
    #         f.write(f"\t_job_buffer[{i}][sel_index]=buffer[{i}]\n")

    #     f.close()

    #     dp = __import__("_dptest", fromlist=["_put"])
    #     self._put = dp._put


def Queue(job_spec: JobSpec = None,
          *,
          capacity: int = 1):
    """
    Returns BaseQueue or FIFOQueue depending on the inputs.
    """

    queue = BaseQueue if job_spec is None else FIFOQueue

    return queue(capacity, job_spec)




import os
import numpy as np
import matplotlib.pyplot as plt
from random import random
from time import time_ns
from pickle import dump

import ptlib._backend as PTCONFIG
from ptlib.core.metadata import MetadataManager


class Diagram:
    """ Parallel Timing Diagram *** come back """

    # spacing between timing curves in the PTD
    PTD_SPACING = PTCONFIG._DIAGRAM.PTD_SPACING

    # width of strings printed to console
    CONSOLE_WIDTH = PTCONFIG._DIAGRAM.CONSOLE_WIDTH

    # time scaling divider
    TIME_DIV = PTCONFIG._DIAGRAM.TIME_DIV

    def __init__(self,
                 total_jobs: int = None,
                 start_time: int = None,
                 finish_time: int = None,
                 process_names: dict[str] = None,
                 metadata: dict[int, int] = None,
                 *,
                 meta_manager: MetadataManager = None):
        """
        Parameters:
            total_jobs -- int
                The number of jobs processed by the pipeline.
            start_time -- int
                The start time of the main process in nanoseconds.
            finish_time: int
                The finish time of the main process in nanoseconds.
            process_names: dict[task_id, worker_id]
                Dict where each entry is the string name for worker with
                `id=worker_id` assigned to the task with `id=task_id`.
            metadata -- dict[task_id, worker_id][i] = (job_start, job_finish)
                Look at `ptlib.core.metadata.MetadataManager` for a more 
                detailed explanation.
        """

        assert isinstance(
            meta_manager, MetadataManager), f"meta_manager is the wrong type: {type(meta_manager)}"

        self._total_jobs = meta_manager._total_jobs if meta_manager else total_jobs
        self._start = meta_manager._start if meta_manager else start_time
        self._finish = meta_manager._finish if meta_manager else finish_time

        # combine names and metadata
        names = meta_manager._names if meta_manager else process_names
        meta = meta_manager._meta if meta_manager else metadata

        self._meta = {index: (names[index], meta[index])
                      for index in names.keys()}

        # create diagram plots
        self.ax_ptd = plt.subplot2grid((2, 3), (0, 0), colspan=5)
        self.ax_jobs = plt.subplot2grid((2, 3), (1, 0), colspan=1)
        self.ax_times = plt.subplot2grid((2, 3), (1, 1), colspan=1)
        self.ax_rates = plt.subplot2grid((2, 3), (1, 2), colspan=1)

    def show(self):
        """ 
        Displays current graphs by using `plt.show()`. 
        """

        return plt.show()

    def graph_all(self, save_path: str = ""):
        """
        Generates each of the three graphs. If `save_path` is empty, then 
        the graphs are shown. Otherwise, the graphs are written to 
        `save_path` as a .pkl file. 
        """

        self.graph_ptd()
        self.graph_stats()
        self.table_stats()

        if save_path == "":
            self.show()
        else:
            path = os.path.join(save_path + ".pkl")
            print(f"Saving Parallel Timing Diagram to: {path}")
            dump(plt.gcf(), open(path, "wb"))

    def graph_ptd(self):
        """ 
        Generates parallel timing diagram. 
        """

        print("Generating PTD Graph...")

        # set finish time if it is None
        self._finish = self._finish or time_ns()

        # the y value of the top most timing curve
        y = self.PTD_SPACING * len(self._meta)
        for name, metadata in self._meta.values():
            # create PTD curves and decrement y position
            if True:
                self._create_ptd_lines(name, metadata, y)
                y -= self.PTD_SPACING

        # set graph labels
        self.ax_ptd.set_ylabel("High/Low")
        self.ax_ptd.set_xlabel("Unix Time (ns)")
        self.ax_ptd.legend()

    def graph_stats(self):
        """ 
        Displays the PTD statistics as a bar graph. 
        """

        print("Generating PTD Stats Bar Graph...")

        stats = self.get_stats(graph=True)
        names, worker_ids, num_jobs, \
            times_on, times_off, rates_on, rates_off = zip(*stats)

        x = np.arange(len(names))
        width = 0.5

        # jobs
        jobs_bar = self.ax_jobs.bar(x, num_jobs, width)
        self.ax_jobs.bar_label(jobs_bar, padding=2)
        self.ax_jobs.set_ylabel("n")
        self.ax_jobs.set_xticks(x)
        self.ax_jobs.set_xticklabels(names, rotation=315)
        self.ax_jobs.set_title("Number of Jobs")

        # times
        times_on = np.around(np.array(times_on) / self.TIME_DIV, decimals=2)
        times_off = np.around(np.array(times_off) / self.TIME_DIV, decimals=2)
        times_on_bar = self.ax_times.bar(
            x - (width - .15) / 2, times_on, (width - .15), label="On")
        times_off_bar = self.ax_times.bar(
            x + (width - .15) / 2, times_off, (width - .15), label="Off")
        self.ax_times.bar_label(times_on_bar, padding=2)
        self.ax_times.bar_label(times_off_bar, padding=2)
        self.ax_times.set_ylabel("Time (s)")
        self.ax_times.set_xticks(x)
        self.ax_times.set_xticklabels(names, rotation=315)
        self.ax_times.set_title("Time Spent")
        self.ax_times.legend()

        # rates
        rates_on = np.around(np.array(rates_on) * self.TIME_DIV, decimals=2)
        rates_off = np.around(np.array(rates_off) * self.TIME_DIV, decimals=2)
        rates_on_bar = self.ax_rates.bar(
            x - (width - .15) / 2, rates_on, (width - .15), label="On")
        rates_off_bar = self.ax_rates.bar(
            x + (width - .15) / 2, rates_off, (width - .15), label="Off")
        self.ax_rates.bar_label(rates_on_bar, padding=2)
        self.ax_rates.bar_label(rates_off_bar, padding=2)
        self.ax_rates.set_ylabel("Rate (jobs / s)")
        self.ax_rates.set_xticks(x)
        self.ax_rates.set_xticklabels(names, rotation=315)
        self.ax_rates.set_title("Job Rate")
        self.ax_rates.legend()

    def table_stats(self):
        """ 
        Creates a table containing PTD statistics. 
        """

        print("Generating PTD Stats Table...")

    def get_stats(self, graph=False):
        """ 
        Generates and returns PTD statistics. 
        """

        self._finish = self._finish or time_ns()
        stats = list()
        for (task_id, worker_id), (name, metadata) in self._meta.items():

            num_jobs = len(metadata)
            time_on = time_off = 0

            # Start and finish times are specificed by first pair
            assert len(
                metadata) > 0, f"Error: Worker has no metadata stored. | Name: {name}, Metadata: {metadata}"
            off_since, finish = metadata[0]
            metadata = metadata[1:]

            # increment accumulators
            for ont, offt in metadata:
                time_off += ont - off_since
                time_on += offt - ont
                off_since = offt

            # check edge case
            if finish > off_since:
                time_off += finish - off_since

            # calculate rates
            rate_on = num_jobs / time_on if time_on else 0
            rate_off = num_jobs / time_off if time_on else 0

            stats.append((name, worker_id, num_jobs, time_on,
                          time_off, rate_on, rate_off))

        return stats

    def _create_ptd_lines(self, name: str, metadata: list, y: float):
        """ 
        Creates curve on the timing diagram for a single worker.
        """

        # Start and finish times are specificed by first pair
        assert len(
            metadata) > 0, f"Error: PTProcess has no metadata stored. | Name: {name}, Metadata: {metadata}"

        # set inital on/off times
        off_since, finish = metadata[0]
        metadata = metadata[1:]

        # give curve a random color
        color = (random(), random(), random())

        # add label
        self.ax_ptd.hlines(y, off_since, off_since, color=color, label=name)

        # create vertical and horizontal lines
        for ont, offt in metadata:
            self.ax_ptd.hlines(y, off_since, ont, color=color)
            self.ax_ptd.vlines(ont, y, y + 1, color=color)
            self.ax_ptd.hlines(y + 1, ont, offt, color=color)
            self.ax_ptd.vlines(offt, y, y + 1, color=color)
            off_since = offt

        # extend LOW line
        if finish > off_since:
            self.ax_ptd.hlines(y, off_since, finish, color=color)

    def __str__(self):
        """ Formats the timing diagram statistics in a readable way. """

        s = "=" * self.CONSOLE_WIDTH + "\n"

        # print overall performance
        runtime = (self._finish - self._start) / self.TIME_DIV
        jps = self._total_jobs / runtime
        s += "\n*** OVERALL PERFORMANCE: "
        s += f" | Pipeline Runtime: {runtime} s"
        s += f" | Processed {self._total_jobs} jobs @ {jps:.2f} j/s ***\n\n"

        # print job specific performance
        for name, worker_id, num_jobs, time_on, time_off, rate_on, rate_off in self.get_stats():
            stat_array = [time_on / self.TIME_DIV, time_off / self.TIME_DIV,
                          rate_on * self.TIME_DIV, rate_off * self.TIME_DIV]
            time_on, time_off, rate_on, rate_off = np.around(
                stat_array, decimals=2)
            s += f"Task: {name} & Worker ID: {worker_id} -- {num_jobs} jobs"
            s += f" | {time_on} s on, {time_off} s off"
            s += f" | {rate_on} j/s on, {rate_off} j/s off\n"

        s += "\n" + "=" * self.CONSOLE_WIDTH

        return s



from time import time_ns

#################### BACKEND CONFIGURATION ####################
"""
This file contains backend configurations for several objects, where 
each class corresponds to the name of a file script in the package. 
"""


class _METADATA:
    JOB_SPEC_EXAMPLES = [[1, 1, 1], [time_ns(), time_ns()]]
    QUEUE_MAX_SIZE = 30


class _DIAGRAM:
    PTD_SPACING = 1.5
    CONSOLE_WIDTH = 100
    TIME_DIV = 1e9          # Nanoseconds
###############################################################



from time import time_ns

import ptlib._backend as ptconfig
from ptlib.core.job import JobSpec
from ptlib.core.queue import BaseQueue, Queue


class MetadataManager:
    """
    Class to abstract metadata initialization and updates. Metadata is 
    transported through a `ptlib.core.queue.FIFOQueue` as if it were a job.

    Attributes:
        metadata_spec -- ptlib.core.job.JobSpec
            The specification for the metadata queue which has two jobs. The 
            first is used to identify the worker and has the structure
            [task id, worker id, S/F indicator], where the S/F indicator is 1 if
            the incoming data is for the start/finish of the worker process, 
            and 0 if the incoming data is for the worker's job. The second job 
            transmitted through the metadata queue is the metadata itself and 
            has the structure
            [start time, finish time], where each time is in nanoseconds.
        meta_q_cap -- int
            The capacity of the metadata queue. 
        _total_jobs -- int
            The number of jobs processed by the pipeline.
        _start -- int
            The start time in nanoseconds of the overall pipeline (the time at 
            which the `run` function of the controller was called). 
        _finish -- int
            The finish time in nanoseconds of the overall pipeline (the time at 
            which the `run` function of the controller terminated. 
        _names -- dict
            Dictionary of worker names where 
            `_names[task id, worker id]=worker name`. 
        _meta -- dict
            Dictionary of metadata where 
            `_meta[task id, worker id][i]=(start time, finish time)`. Note that 
            each start and finish time are in nanoseconds. If `i==0`, then 
            the metadata corresponds to the start or finish of the worker 
            process. Otherwise, the metadata corresponds to the ith job.
        """

    # job specification for metadata (task id, worker id, if data is start/finish time)
    metadata_spec = JobSpec(name="id", example=ptconfig._METADATA.JOB_SPEC_EXAMPLES[0]) + \
        JobSpec(name="data", example=ptconfig._METADATA.JOB_SPEC_EXAMPLES[1])

    # the queue max size
    meta_q_cap = ptconfig._METADATA.QUEUE_MAX_SIZE

    def __init__(self, pipeline, total_jobs=None):
        # store total number of jobs processed by the pipeline
        self._total_jobs = total_jobs

        # initialize start and finish times
        self._start = self._finish = None

        # create mappings for worker names and metadata
        self._names, self._meta = dict(), dict()

        # create initial entries
        for task in pipeline.iter_tasks():
            for worker_id in range(task.num_workers):
                # set name
                self._names[task.id,
                            worker_id] = f"{task.name}: {worker_id}"

                # set metadata to list with one element (will get overridden in first `update` call)
                self._meta[task.id, worker_id] = [None]

        # queue for retreiving metadata from asnychronous workers
        self.meta_q = Queue(self.metadata_spec,
                            capacity=self.meta_q_cap)

        # link metadata queue and get local buffer
        meta_buffer = self.meta_q._link_mem()
        self._id_buffer = meta_buffer["id"]
        self._data_buffer = meta_buffer["data"]

    def update(self):
        """
        Update metadata while `self.meta_q` is not empty.
        """

        while self.meta_q.get() is not BaseQueue.Empty:
            # get the index and new metadata
            *index, sf_ind = self._id_buffer
            new_meta = tuple(self._data_buffer)

            # if the S/F indicator is HIGH, set the first index
            if sf_ind == 1:
                self._meta[tuple(index)][0] = new_meta

            # otherwise add job metadata to list
            else:
                self._meta[tuple(index)].append(new_meta)

    def set_time(self):
        """ 
        Used to set main pipeline start and finish times. The first call of 
        tihs function will set the start time, and all subsequent calls will 
        update the finish time. 
        """

        if self._start is None:
            self._start = time_ns()
        else:
            self._finish = time_ns()



import multiprocessing as mp

from ptlib.core.metadata import MetadataManager
from ptlib.core.task import Task, EmptyTask
from ptlib.core.queue import Queue
from ptlib.utils.diagram import Diagram


class Controller:
    """ The controller. *** COME BACK *** """

    def __init__(self,
                 pipeline: Task,
                 queue_max_size: int = 5,
                 total_jobs: int = None):
        """
        Parameters:
            pipeline: ptlib.Task
                Task or pipeline of tasks connected by `>>` operator.
            queue_max_size: (optional) int
                The the maximum number of jobs that can be stored in a queue.
            total_jobs: (optional) int
                The number of jobs being processed by the pipeline. Used for 
                runtime analytics, not computation. Can be passed as an 
                argument or set by overloading the `Task.get_total_jobs` 
                function in the first task of `pipeline`.
        """

        # for process start method to spawn correctly
        mp.set_start_method("spawn", force=True)

        # get the number of jobs intended to be processed by the pipeline
        total_jobs = total_jobs or pipeline.get_total_jobs()

        # create metadata manager before tasks are set up
        self.meta_manager = MetadataManager(pipeline, total_jobs)

        # store initialization arguments
        self.pipeline = pipeline
        self.queue_max_size = queue_max_size

        # set up tasks
        self._set_up_tasks()

    def run(self):
        """ 
        The main run loop for the pipeline. 
        """

        # set start time
        self.meta_manager.set_time()

        # start all worker processes
        for task in self.pipeline.iter_tasks():
            task._start_workers()

        task = self.pipeline
        while task is not EmptyTask:
            # update metadata
            self.meta_manager.update()

            # if current task finishes, send kill signal to workers
            if not task._workers_running():
                print(f"Task Finished: {task.name}")
                task = task.next
                task._kill_workers()

        # finish retreiving metadata (in case loop exits before getting all metadata)
        self.meta_manager.update()

        # set finish time
        self.meta_manager.set_time()

        print(self.meta_manager._meta)
        print("controller done")

    def graph(self, save_path=""):
        """
        Creates and shows parallel timing diagram. If `save_path` is empty, then 
        the graphs are shown. Otherwise, the graphs are written to 
        `save_path` as a .pkl file. 
        """

        diag = Diagram(meta_manager=self.meta_manager)
        print(diag)
        diag.graph_all(save_path)

    def _set_up_tasks(self):
        """
        Sets up each task with the appropriate queue.
        """

        # set default input job and fake input queue
        input_job, input_q, = None, Queue()

        for task in self.pipeline.iter_tasks():
            print(f"Creating Task: {task.name} | ID: {task.id}")

            # skip last task
            if task.next is EmptyTask:
                break

            # try to infer output job structure and set output to input of next task
            # input_job, job_specs = task._infer_structure(input_job)
            job_specs, input_job = task._infer_structure(input_job)

            # create and store output queue
            output_q = Queue(job_specs, capacity=self.queue_max_size)

            # set input queue of task (see method documentation for reason)
            task._set_input_queue(input_q)

            # create workers and assign them to task
            task.create_workers(input_q, output_q,
                                self.meta_manager.meta_q)

            # set input queue of task.next to output queue of task
            input_q = output_q

            # update task
            task = task.next

        # like above
        task._set_input_queue(input_q)

        # create workers and assign them to the final task
        task.create_workers(input_q, Queue(), self.meta_manager.meta_q)

    def _add_worker(self, name, task_id, worker_id):
        """
        Called when ptlib.core.worker.Worker is initialized. Registers the
        worker and shared data structures.
        """

        # stores pairs of timestamps of when job starts and finishes
        pairs = self.list()

        # latest job start time
        latest_start = self.Value("i", 0)

        self._worker_map[task_id, worker_id] = (name, pairs, latest_start)



from multiprocessing import Process
from time import time_ns

from ptlib.core.queue import BaseQueue


class Worker(Process):
    """ Worker class. *** come back*** """

    def __init__(self, Task, num_workers, task_id, worker_id, input_q, output_q, meta_q):
        self.EXIT_FLAG = False
        self.num_workers = num_workers
        self.task_id = task_id
        self.id = worker_id

        # create worker process
        super().__init__(target=self.work,
                         args=(Task, input_q,
                               output_q, meta_q),
                         daemon=True)

    def work(self, Task, input_q, output_q, meta_q):
        """
        The main processing loop for `task`.
        """

        # create task object and set correct task id
        task = Task(num_workers=self.num_workers, task_id=self.task_id)

        # link queues to memory
        input_job = input_q._link_mem()
        output_job = output_q._link_mem()
        meta_buffer = meta_q._link_mem()

        # set up local metadata buffers
        id_buffer = meta_buffer["id"]
        data_buffer = meta_buffer["data"]

        # set the first job in buffer to (task id, worker id, and S/F indicator to HIGH)
        id_buffer[:] = [task.id, self.id, 1][:]

        # give main thread worker start time
        start_time = time_ns()
        data_buffer[:] = [start_time, start_time][:]
        # (if metadata seems off, add while loop)
        assert meta_q.put() is not BaseQueue.Full, "meta q should always put (prejob)"

        # set S/F indicator to LOW
        id_buffer[-1] = 0

        # create job mapping
        job_map = task.create_map(self, input_job, output_job)

        input_status = BaseQueue.Empty

        while not self.EXIT_FLAG:
            # t = time_ns()
            if (input_status := input_q.get()) is BaseQueue.Empty:
                continue
            elif input_status is BaseQueue.Closed:
                break
            # print(f"Task: {task.id} | Get Time: {time_ns() - t}")

            # record start time
            data_buffer[0] = time_ns()

            # map input job to compute output job
            job_map()

            # print(output_job_buf)
            # if output_job[0] is not None:
            #     for i in range(len(output_job)):
            #         print(output_job_buf[i].shape, output_job[i].shape)
            #         output_job_buf[i][:] = output_job[i]

            # record finish time
            data_buffer[1] = time_ns()

            # put metadata into queue (if metadata seems off, add while loop)
            assert meta_q.put() is not BaseQueue.Full, "meta q should always put"

            t = time_ns()
            while output_q.put() is BaseQueue.Full:
                pass
            # print(f"Task: {task.id} | Put Time: {(time_ns() - t)/1e9}")

        # run cleanup routine
        task.cleanup()

        # record finish time
        finish_time = time_ns()

        # set S/F indicator to HIGH
        id_buffer[-1] = 1

        # give main thread worker finish time
        data_buffer[:] = [start_time, finish_time][:]
        meta_q.put()  # (if metadata seems off, add while loop)

        print(f"Worker Done -- Task: {task.name} | ID: {self.id}")



from typing import Iterable
import numpy as np

from ptlib.core.job import Job, Job
from ptlib.errors import WorkerCreationError, WorkerStartError
from ptlib.core.queue import Queue
from ptlib.core.worker import Worker


class EmptyTask:
    def _kill_workers():
        pass


class Task:
    """
    Template class for making new Task objects.

    Parameters:
        config -- ptlib.task.config
            Task configuaration object used to set up task at runtime.

    Attributes:
        config -- ptlib.task.config
            Task configuration object passed at runtime.
        EXIT_FLAG -- bool
            Used to determine loop termination in process.
    """

    class Exit:
        pass

    # ---------------------------------------------------------------------- #
    # Constructors

    def __init__(self, num_workers: int = 1, task_id: int = 0):
        """
        When subclassing `Task`, do not implement a different init, or if you 
        do, don't change the arguments.
        """

        self.num_workers = num_workers
        self.id = task_id

        self.workers = list()
        self.next = EmptyTask
        self.input_q = Queue()

    @property
    def ttype(self):
        """
        The task type. 
        """

        return self.__class__

    @property
    def name(self):
        """
        Formats the name of the task instance by looking at the class
        definition.
        """

        return self.ttype.__name__

    # ---------------------------------------------------------------------- #
    # Methods that can be Overloaded by Developer

    def create_map(self, worker: Worker, input_job, output_job):
        """
        Returns a function that maps input job to output job for a specific
        worker.

        The returned function `map_job` should evaluate an expression
        to determine the value of `worker.EXIT_FLAG`. It should also do
        something with each sub job in `input_job: ptlib.core.job.Job` update
        each subjob in `output_job: ptlic.core.job.Job`.

        The default `create_map` should be overloaded.
        """

        def map_job():
            # example exit
            if input_job is Task.Exit:
                worker.EXIT_FLAG = True

            # example modification
            output_job = input_job.copy()
            for i in range(len((input_job))):
                output_job[i][:] = input_job[i][:]

            return output_job

        return map_job

    def get_example_job(self):
        """
        Returns an object resembling the output job of this task. If this is 
        not overloaded by the developer, the controller will try to infer the 
        structure of the output job. 
        """

        return None

    def create_workers(self, input_q, output_q, meta_q):
        """
        Returns a list of pt.Worker objects and stores input queue. Overloading
        this function allows the developer to modify the controller's
        instructions for worker creation.
        """

        if len(self.workers) > 0:
            raise WorkerCreationError(
                f"Workers already exist for task: {self}")

        # create workers
        from copy import copy
        for worker_id in range(self.num_workers):
            self.workers.append(
                Worker(self.ttype, self.num_workers, self.id, worker_id, input_q, output_q, meta_q))

    def get_total_jobs(self):
        """
        Returns the number of jobs processed by the pipeline. This is used for 
        runtime analytics. If needed, this should be overloaded in the first 
        task of the pipeline. 
        """

        return None

    def cleanup(self):
        """
        Routine for task cleanup after termination. This function should be
        overloaded if needed.
        """

        pass

    # ---------------------------------------------------------------------- #
    # Private and Protected Methods

    def _infer_structure(self, input_job):
        """
        Tries to infer the structure (shape, dtype) of task's output job an
        example input job. This is done by creating (and later deleting) a
        temporary worker to analyze the output of `map_job(input_job)`.
        """

        input_job = self.get_example_job() or input_job

        worker = Worker(self.ttype, 0, 0, 0, Queue(), Queue(), Queue())
        output_job = Job()
        self.create_map(worker, input_job, output_job)()
        job_spec, ouput_job = output_job.infer()

        return job_spec, output_job

    def iter_tasks(self):
        """
        Generator that iterates over linked list structure of tasks using the
        `task.next` attribute.
        """

        task = self
        while task is not EmptyTask:
            yield task
            task = task.next

    def _set_input_queue(self, input_q):
        """
        Sets `self.input_q=input_q`. This is important because it guarantees
        that each queue is not accidently garbage collected before it's used.
        Also needed in 'self._kill_workers`.
        """

        self.input_q = input_q

    def _start_workers(self):
        """
        Starts each of the worker processes. This should not be overloaded.
        """

        if len(self.workers) == 0:
            raise WorkerStartError(
                f"Workers have not been created for task: {self}")

        for worker in self.workers:
            worker.start()

    def _workers_running(self):
        """
        Returns True if workers are still running, False if otherwise.
        """

        return any([worker.is_alive() for worker in self.workers])

    def _kill_workers(self):
        """
        Signals workers to exit by closing their input queue.
        """

        self.input_q.close()

    def __rshift__(self, other):
        """
        Allows tasks to be connected into a pipeline using the `>>` operator.
        This operation finds the last task in linked list of tasks by iterating
        over the `task.next` attribute. Assigns `other.id` to the last
        task ID + 1.
        """

        if not isinstance(other, Task):
            return self

        task = self
        while task.next is not EmptyTask:
            task = task.next

        other.id = task.id + 1
        task.next = other

        return self

    # ---------------------------------------------------------------------- #



import numpy as np

from typing import Hashable, Tuple, Any

class JobSpec:
    """ Job specification. *** COME BACK *** """

    def __init__(self,
                 shape: Tuple[int] = None,
                 dtype: np.dtype = None,
                 *,
                 name: Hashable = None,
                 example: np.ndarray = None):
        """
        Parameters:
            shape -- Tuple[int]
                The shape of the data.
            dtype -- np.dtype
                The type of the data.
            name -- Hashable
                The key that will be used to index the `SubJob` in the `Job`. 
            example -- ndarray
                An example data to infer structure and space requirements of
                the data (if provided, shape and dtype will be ignored)
        """

        self.next = None

        # if no shape nor example is given, assumed to be an empty spec
        self._is_empty = shape is None and example is None
        if self._is_empty:
            self.shape = (0,)
            self.dtype = None
            return

        self.name = name
        if example is not None:
            example = np.array(example)
            shape, dtype = example.shape, example.dtype

        self.shape = np.array(shape, dtype=np.ulonglong)
        self.dtype = dtype
        self.nbytes = int(np.nbytes[self.dtype] * np.product(self.shape))

    @classmethod
    def from_output_job(cls, output_job: "Job" or np.ndarray):
        """ 
        Alternative constructor for determing job specifications from an 
        instance of `ptlib.core.job.Job`. 
        """

        job_spec_list = [cls(example=job) for job in output_job]

        job_specs = job_spec_list.pop(0)
        for next_job_spec in job_spec_list:
            job_specs += next_job_spec

        return job_specs

    def get_nbytes(self, capacity: int = 1):
        """
        Calculates the number of bytes required to allocate a `ptlib.Job` 
        specified by `self` in a `ptlib.Queue` of size `capacity`. 
        """

        return capacity * sum([js.nbytes for js in self])

    def __iter__(self):
        js = self
        while js is not None:
            yield js
            js = js.next

    def __add__(self, other_js: "JobSpec") -> "JobSpec":
        """
        LinkedList implementation allows for JobSpec objects to be combined 
        using the addition operator `+`.
        """

        if not isinstance(other_js, JobSpec):
            return self

        if self._is_empty:
            return other_js

        js = self
        while js.next is not None:
            js = js.next

        js.next = other_js

        return self

    def __repr__(self):
        return "[" + str(self) + "]"

    def __str__(self):
        s = " | ".join(
            [f"Name: {js.name}, Shape: {js.shape}, DType: {js.dtype}" for js in self])

        return "[" + s + "]"


class Job(dict):
    """ 
    The Job object. Can be used to infer structure of jobs and as a job 
    itself.
    """

    def __init__(self, job_spec: JobSpec = None):
        super().__init__()

        # if no job spec is passed, then this instance is used for inferance
        if job_spec is None:
            # if dictionary is set-sliced with object v, v is stored in this
            self._subjob = None
            return

        # create local buffer for entire job
        self._buffer = np.ndarray(job_spec.get_nbytes(), dtype=np.int8)

        # create subjobs
        offset = 0
        for js in job_spec:
            self[js.name] = np.ndarray(shape=js.shape,
                                       dtype=js.dtype,
                                       buffer=self._buffer,
                                       offset=offset)

            offset += js.nbytes

    def infer(self):
        """
        Infers the structure of each `SubJob` and returns a `JobSpec` object.
        Calling this function also collapses each embedded `Job` object into 
        a `SubJob`.
        """

        job_spec = JobSpec()
        for key, subjob in self.items():
            if isinstance(subjob, Job):
                subjob = self[key] = subjob.collapse()

            job_spec = job_spec + JobSpec(name=key, example=subjob)

        return job_spec, self

    def collapse(self):
        """
        Recursively collapses embedded `Job` objects into a single numpy 
        array.
        """

        if self._subjob is not None:
            return self._subjob

        subjobs = list()
        for subjob in self.values():
            if isinstance(subjob, Job):
                subjob = subjob.collapse()

            subjobs.append(subjob)

        return np.array(subjobs)

    def __getitem__(self, k):
        # print(k)
        if k not in self:
            self[k] = Job()

        return super().__getitem__(k)

    def __setitem__(self, k, v) -> None:
        # print(k, v)
        # if `k` is a slice, then
        if not isinstance(k, slice):
            return super().__setitem__(k, v)

        self._subjob = v

    def __getattribute__(self, __name: str):
        """
        Overloaded to ensure that inferencing of job structure is not stopped 
        due to a method call that does not pertain to the structure. For 
        example, if `subjob1.fill(0)` is called (where `fill` is a method of a 
        `numpy.ndarray`), we want to continue to infer the structure of the 
        subjob without throwing an error. 
        """

        try:
            return super().__getattribute__(__name)
        except AttributeError:
            return Job()

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        """
        Overloaded to protect runtime interruption during structure inference. 
        """

        try:
            return super().__call__(*args, **kwds)
        except TypeError and AttributeError:
            return None




class WorkerCreationError(Exception):
    """ 
    Error raised when workers are unable to be created for a task.
    """


class WorkerStartError(Exception):
    """
    Error raised when worker processed are unable to be started for a task. 
    """



from ptlib.core.job import JobSpec, Job
from ptlib.core.queue import Queue
from ptlib.core.task import EmptyTask, Task
from ptlib.core.worker import Worker
from ptlib.core.controller import Controller



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ptlib",
    version="0.0.5",
    author="Joseph Whelan",
    author_email="",
    description="This library aims to provide efficient implementation and detaile debugging of solutions to problems in the asynchronous domain.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jfw225/ptlib",
    project_urls={
        "Bug Tracker": "https://github.com/jfw225/ptlib/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=[
        "numpy>=1.19.5",
        "matplotlib>=3.4.3"
    ],
    python_requires=">=3.8",
)